import unittest
from unittest import mock
from datetime import datetime
from singer.catalog import Catalog
import pytz
from tap_mixpanel import main

TEST_CONFIG = {
    "start_date": "2022-09-01",
    "project_timezone": "US/Pacific",
    "api_secret": "API_SECRET",
    "attribution_window": 5,
    "user_agent": "USER_AGENT"
}


class MockArgs:
    """Mock args object class"""

    def __init__(self, config=None, catalog=None, state={}, discover=False) -> None:
        self.config = config
        self.catalog = catalog
        self.state = state
        self.discover = discover


# Setting now time to fix value
@mock.patch("singer.utils.now", return_value=datetime(year=2022, month=10, day=1).replace(tzinfo=pytz.UTC))
@mock.patch("tap_mixpanel.MixpanelClient.check_access")
@mock.patch("singer.utils.parse_args")
@mock.patch("tap_mixpanel._sync")
class TestSyncMode(unittest.TestCase):
    """
    Test the main function for sync mode.
    """

    mock_catalog = {"streams": [
        {"stream": "invoices", "schema": {}, "metadata": {}}]}

    @mock.patch("tap_mixpanel._discover")
    @mock.patch("tap_mixpanel.LOGGER.warning")
    def test_sync_with_catalog(self, mock_logger, mock_discover, mock_sync,
                               mock_args, mock_check_access, mock_now):
        """
        Test sync mode with catalog given in args.
        """
        test_config = {**TEST_CONFIG, "start_date": "2020-09-01"}

        mock_args.return_value = MockArgs(config=test_config,
                                          catalog=Catalog.from_dict(self.mock_catalog))
        main()

        # Verify `_sync` is called with expected arguments
        # Start_date will be set as 1 year from now date-time(2022-10-01)
        mock_sync.assert_called_with(client=mock.ANY,
                                     config=test_config,
                                     catalog=Catalog.from_dict(self.mock_catalog),
                                     state={},
                                     start_date="2021-10-01T00:00:00.000000Z")

        # verify `_discover` function is not called
        self.assertFalse(mock_discover.called)

        # Verify that for start_date before 1 year given warning logger is printed
        mock_logger.assert_any_call("start_date greater than 1 year maximum for API.")

    @mock.patch("tap_mixpanel._discover")
    def test_without_catalog(self, mock_discover, mock_sync, mock_args, mock_check_access, mock_now):
        """
        Test sync mode without catalog given in args.
        """

        discover_catalog = Catalog.from_dict(self.mock_catalog)
        mock_discover.return_value = discover_catalog
        mock_args.return_value = MockArgs(config=TEST_CONFIG)
        main()

        # verify `_discover` and `_sync` function is called
        self.assertTrue(mock_discover.called)
        self.assertTrue(mock_sync.called)

    def test_sync_with_state(self, mock_sync, mock_args, mock_check_access, mock_now):
        """
        Test sync mode with the state given in args.
        """
        mock_state = {"bookmarks": {"projec ts": ""}}
        mock_args.return_value = MockArgs(config=TEST_CONFIG,
                                          catalog=Catalog.from_dict(self.mock_catalog),
                                          state=mock_state)
        main()

        # Verify `_sync` is called with expected arguments
        mock_sync.assert_called_with(client=mock.ANY,
                                     config=TEST_CONFIG,
                                     catalog=Catalog.from_dict(self.mock_catalog),
                                     state=mock_state,
                                     start_date=TEST_CONFIG["start_date"])

    @mock.patch("tap_mixpanel._discover")
    def test_discover_mode(self, mock_discover, mock_sync, mock_args, mock_check_access, mock_now):
        """
        Test discover mode calls discover function to generate catalog.
        """

        discover_catalog = Catalog.from_dict(self.mock_catalog)
        mock_discover.return_value = discover_catalog
        mock_args.return_value = MockArgs(config=TEST_CONFIG,
                                          discover=True)
        main()

        # Verify `_discover` is called
        self.assertTrue(mock_discover.called)



import unittest
from unittest import mock
from parameterized import parameterized
from singer import Catalog
from tap_mixpanel.sync import write_schemas_recursive, get_streams_to_sync, sync


def get_stream_catalog(stream_name, is_selected=False):
    """Return catalog for stream"""
    return {
        "schema": {},
        "tap_stream_id": stream_name,
        "stream": stream_name,
        "metadata": [
            {
                "breadcrumb": [],
                "metadata":{"selected": is_selected}
            }
        ],
        "key_properties": []
    }


def get_catalog(parent=False, child=False):
    """Return complete catalog"""

    return Catalog.from_dict({
        "streams": [
            get_stream_catalog("engage"),
            get_stream_catalog("cohorts", parent),
            get_stream_catalog("cohort_members", child),
            get_stream_catalog("revenue", parent),
        ]
    })


class TestSyncFunctions(unittest.TestCase):
    """
    Test `sync` function.
    """

    @parameterized.expand([
        # ["test_name", "mock_catalog", "selected_streams", "synced_streams"]
        ["only_parent_selected", get_catalog(parent=True), ["cohorts", "revenue"], 2],
        ["only_child_selected", get_catalog(child=True), ["cohort_members"], 1],
        ["both_selected", get_catalog(parent=True, child=True), ["cohorts", "cohort_members", "revenue"], 2],
        ["No_streams_selected", get_catalog(), [], 0],
    ])
    @mock.patch("singer.write_state")
    @mock.patch("singer.write_schema")
    @mock.patch("tap_mixpanel.streams.MixPanel.sync")
    def test_sync(self, test_name, mock_catalog, selected_streams, synced_streams,
                  mock_sync_endpoint, mock_write_schemas, mock_write_state):
        """
        Test sync function.
        """
        client = mock.Mock()
        sync(client=client, config={}, state={}, catalog=mock_catalog, start_date="START_DATE")

        # Verify write schema is called for selected streams
        self.assertEqual(mock_write_schemas.call_count, len(selected_streams))
        for stream in selected_streams:
            mock_write_schemas.assert_any_call(stream, mock.ANY, mock.ANY)

        # Verify sync object was called for syncing parent streams
        self.assertEqual(mock_sync_endpoint.call_count, synced_streams)


class TestGetStreamsToSync(unittest.TestCase):
    """
    Testcase for `get_stream_to_sync` in sync.
    """

    @parameterized.expand([
        # ["test_name", "selected_streams", "expected_streams"]
        ['test_parent_selected', ["funnels", "cohorts"], ["funnels", "cohorts"]],
        ['test_child_selected', ["cohort_members"], ["cohorts"]],
        ['test_both_selected', ["cohorts", "cohort_members", "revenue"], ["cohorts", "revenue"]]
    ])
    def test_sync_streams(self, test_name, selected_streams, expected_streams):
        """
        Test that if an only child is selected in the catalog,
        then `get_stream_to_sync` returns the parent streams if selected stream is child.
        """
        sync_streams = get_streams_to_sync(selected_streams)

        # Verify that the expected list of streams is returned
        self.assertEqual(sync_streams, expected_streams)


@mock.patch("singer.write_schema")
class TestWriteSchemas(unittest.TestCase):
    """
    Test `write_schemas` function.
    """

    test_catalog = Catalog.from_dict({"streams": [
        get_stream_catalog("cohorts"),
        get_stream_catalog("cohort_members"),
        get_stream_catalog("revenue")
    ]})

    @parameterized.expand([
        # ["test_name", "selected_streams", "mock_write_schema"]
        ["parents_selected", ["cohorts"]],
        ["child_selected", ["cohort_members"]],
        ["parent_and_child_selected", ["cohorts", "cohort_members"]],
    ])
    def test_write_schema(self, mock_write_schema, test_name, selected_streams):
        """
        Test that only schema is written for only selected streams.
        """
        write_schemas_recursive("cohorts", self.test_catalog, selected_streams)
        # Verify write_schema function is called for only selected streams.
        self.assertEqual(mock_write_schema.call_count, len(selected_streams))
        for stream in selected_streams:
            mock_write_schema.assert_any_call(stream, mock.ANY, mock.ANY)

    @mock.patch("tap_mixpanel.streams.LOGGER.error")
    def test_OS_exception(self, mock_logger, mock_schema):
        """
        Test on occurrence of OS error, Expected error logger is printed.
        """
        mock_schema.side_effect = OSError()
        with self.assertRaises(OSError):
            write_schemas_recursive("cohorts", self.test_catalog, ["cohorts"])

        # Verify expected logger is printed
        mock_logger.assert_called_with("OS Error writing schema for: %s", "cohorts")



#!/usr/bin/env python3

import json
import sys
from datetime import timedelta

import singer
from singer import utils
from singer.utils import strftime, strptime_to_utc

from tap_mixpanel.client import MixpanelClient
from tap_mixpanel.discover import discover as _discover
from tap_mixpanel.sync import sync as _sync

LOGGER = singer.get_logger()

REQUEST_TIMEOUT = 300
REQUIRED_CONFIG_KEYS = [
    "project_timezone",
    "api_secret",
    "attribution_window",
    "start_date",
    "user_agent",
]


def do_discover(client, properties_flag):
    """Call the discovery function.

    Args:
        client (MixpanelClient): Client object to make http calls.
        properties_flag (str): Setting this argument to `true` ensures that new properties on
                               events and engage records are captured.
    """
    LOGGER.info("Starting discover")
    catalog = _discover(client, properties_flag)
    json.dump(catalog.to_dict(), sys.stdout, indent=2)
    LOGGER.info("Finished discover")


@singer.utils.handle_top_exception(LOGGER)
def main():
    """
    Run discover mode or sync mode.
    """
    parsed_args = singer.utils.parse_args(REQUIRED_CONFIG_KEYS)

    start_date = parsed_args.config["start_date"]
    # Set request timeout to config param `request_timeout` value.
    # If value is 0, "0", "" or not passed then it sets default to 300 seconds.
    config_request_timeout = parsed_args.config.get("request_timeout")
    if config_request_timeout and float(config_request_timeout):
        request_timeout = float(config_request_timeout)
    else:
        request_timeout = REQUEST_TIMEOUT

    start_dttm = strptime_to_utc(start_date)
    now_dttm = utils.now()
    if parsed_args.config.get("end_date"):
        now_dttm = strptime_to_utc(parsed_args.config.get("end_date"))
    delta_days = (now_dttm - start_dttm).days
    if delta_days >= 365:
        delta_days = 365
        start_date = strftime(now_dttm - timedelta(days=delta_days))
        LOGGER.warning("start_date greater than 1 year maximum for API.")
        LOGGER.warning("Setting start_date to 1 year ago, %s", start_date)

    # Check support for EU endpoints
    if str(parsed_args.config.get("eu_residency")).lower() == "true":
        api_domain = "eu.mixpanel.com"
    else:
        api_domain = "mixpanel.com"

    with MixpanelClient(
        parsed_args.config["api_secret"],
        api_domain,
        request_timeout,
        parsed_args.config["user_agent"],
    ) as client:

        state = {}
        if parsed_args.state:
            state = parsed_args.state

        config = parsed_args.config
        client.__api_domain = api_domain
        properties_flag = config.get("select_properties_by_default")

        if parsed_args.discover:
            do_discover(client, properties_flag)
        else:
            catalog = parsed_args.catalog
            if not catalog:
                catalog = _discover(client, properties_flag)
            _sync(
                client=client,
                config=config,
                catalog=catalog,
                state=state,
                start_date=start_date,
            )


if __name__ == "__main__":
    main()



import unittest
import requests

from unittest import mock
from parameterized import parameterized

from tap_mixpanel import client

# Mock response
REQUEST_TIMEOUT = 300


class MockResponse:
    """Mocked standard HTTPResponse to test error handling."""

    def __init__(
        self, status_code, resp = "", content=[""], headers=None, raise_error=True, text={}
    ):
        self.json_data = resp
        self.status_code = status_code
        self.content = content
        self.headers = headers
        self.raise_error = raise_error
        self.text = text
        self.reason = "error"

    def raise_for_status(self):
        """If an error occur, this method returns a HTTPError object.

        Raises:
            requests.HTTPError: Mock http error.

        Returns:
            int: Returns status code if not error occurred.
        """
        if not self.raise_error:
            return self.status_code

        raise requests.HTTPError("mock sample message")

    def json(self):
        """Returns a JSON object of the result."""
        return self.text


def get_mock_http_response(content, status_code):
    """Mock response for timezone related error messages.

    Args:
        content (str): Returns the content of the response, in bytes.
        status_code (int): Returns a number that indicates the status.

    Returns:
        request.Response: Custom mock response.
    """
    response = requests.Response()
    response.status_code = status_code
    response.headers = {}
    response._content = content.encode()
    return response


# Mock time.sleep to reduce the time
@mock.patch("time.sleep", return_value=None)
class TestMixpanelErrorHandling(unittest.TestCase):
    """
    Test case to verify the custom error message and
    back off is implemented for mentioned errors in tests.
    """

    timeout_400_error = {
        "request": "/api/2.0/engage/revenue?from_date=2020-02-01&to_date=2020-03-01",
        "error": "Timeout Error.",
    }

    def mock_400_different_timezone(*args, **kwargs):
        """Mock 400 error response with with different timezone.

        Returns:
            requests.Response: Returns mock 400 error response.
        """
        content = " to_date cannot be later than today"
        return get_mock_http_response(content, 400)

    def mock_send_error(*args, **kwargs):
        """Mock error response with description in \'error\' field.

        Returns:
            requests.Response: Returns mock 404 error response.
        """
        content = '{"error": "Resource not found error message from API response field \'error\'."}'
        return get_mock_http_response(content, 404)

    def mock_send_message(*args, **kwargs):
        """Mock error response with description in \'message\' field.

        Returns:
            requests.Response: Returns mock 404 error response.
        """
        content = '{"message": "Resource not found error message from API response field \'message\'."}'
        return get_mock_http_response(content, 404)

    @parameterized.expand([
        ["400 error", 400, MockResponse(400), client.MixpanelBadRequestError, "A validation exception has occurred.(Please verify your credentials.)"],
        ["400 different timezone error", 400, mock_400_different_timezone(), client.MixpanelBadRequestError, "A validation exception has occurred. Please validate the timezone with the MixPanel UI under project settings."],
        ["400 timeout error", 400, MockResponse(400, text=timeout_400_error), client.MixpanelBadRequestError, "Timeout Error.(Please verify your credentials.)"],
        ["401 error", 401, MockResponse(401), client.MixpanelUnauthorizedError, "Invalid authorization credentials."],
        ["402 error", 402, MockResponse(402), client.MixpanelPaymentRequiredError, "Your current plan does not allow API calls. Payment is required to complete the operation."],
        ["403 error", 403, MockResponse(403), client.MixpanelForbiddenError, "User does not have permission to access the resource."],
        ["404 error", 404, MockResponse(404), client.MixpanelNotFoundError, "The resource you have specified cannot be found."],
        ["404 error", 404, mock_send_error(), client.MixpanelNotFoundError, "Resource not found error message from API response field 'error'."],
        ["404 error", 404, mock_send_message(), client.MixpanelNotFoundError, "Resource not found error message from API response field 'message'."],
        ["429 error", 429, MockResponse(429), client.Server429Error, "The API rate limit for your organization/application pairing has been exceeded."],
    ])
    @mock.patch("requests.Session.request")
    def test_perform_request_exception_handling(
        self, test_name, error_code, mock_response, error, error_message, mock_request, mock_sleep,
    ):
        """
        Test that `perform_request` method handle error with proper message.
        """
        mock_request.return_value = mock_response
        mock_client = client.MixpanelClient(
            api_secret="mock_api_secret",
            api_domain="mock_api_domain",
            request_timeout=REQUEST_TIMEOUT,
        )
        with self.assertRaises(error) as e:
            mock_client.perform_request("GET")

        expected_error_message = (
            f"HTTP-error-code: {error_code}, Error: {error_message}"
        )

        # Verifying the message formed for the custom exception
        self.assertEqual(str(e.exception), expected_error_message)

    @parameterized.expand([
        ["400 error", 400, MockResponse(400), client.MixpanelBadRequestError, "A validation exception has occurred.(Please verify your credentials.)"],
        ["400 different timezone error", 400, mock_400_different_timezone(), client.MixpanelBadRequestError, "A validation exception has occurred. Please validate the timezone with the MixPanel UI under project settings."],
        ["400 timeout error", 400, MockResponse(400, text=timeout_400_error), client.MixpanelBadRequestError, "Timeout Error.(Please verify your credentials.)"],
        ["401 error", 401, MockResponse(401), client.MixpanelUnauthorizedError, "Invalid authorization credentials."],
        ["403 error", 403, MockResponse(403), client.MixpanelForbiddenError, "User does not have permission to access the resource."],
        ["404 error", 404, MockResponse(404), client.MixpanelNotFoundError, "The resource you have specified cannot be found."],
        ["404 error", 404, mock_send_error(), client.MixpanelNotFoundError, "Resource not found error message from API response field 'error'."],
        ["404 error", 404, mock_send_message(), client.MixpanelNotFoundError, "Resource not found error message from API response field 'message'."],
        ["429 error", 429, MockResponse(429), client.Server429Error, "The API rate limit for your organization/application pairing has been exceeded."],
        ["500 error", 500, MockResponse(500), client.MixpanelInternalServiceError, "Server encountered an unexpected condition that prevented it from fulfilling the request."],
        ["501 error", 501, MockResponse(501), client.MixpanelError, "Unknown Error"],
    ])
    @mock.patch("requests.Session.get")
    def test_check_access_exception_handling(
        self, test_name, error_code, mock_response, error, error_message, mock_request, mock_sleep,
    ):
        """
        Test that `check_access` method handle error with proper message.
        """
        mock_request.return_value = mock_response
        mock_client = client.MixpanelClient(
            api_secret="mock_api_secret",
            api_domain="mock_api_domain",
            request_timeout=REQUEST_TIMEOUT,
        )
        with self.assertRaises(error) as e:
            mock_client.check_access()

        expected_error_message = (
            f"HTTP-error-code: {error_code}, Error: {error_message}"
        )

        # Verifying the message formed for the custom exception
        self.assertEqual(str(e.exception), expected_error_message)

    @parameterized.expand(
        [
            ["500 error", MockResponse(500), client.MixpanelInternalServiceError],
            ["501 error", MockResponse(501), client.Server5xxError],
        ]
    )
    @mock.patch("requests.Session.request")
    def test_request_with_handling_for_5xx_exception_handling(
        self, test_name, mock_response, error, mock_request, mock_sleep
    ):
        """
        Test that `perform_request` method handle 5xx error with proper message.
        """
        mock_request.return_value = mock_response
        mock_client = client.MixpanelClient(
            api_secret="mock_api_secret",
            api_domain="mock_api_domain",
            request_timeout=REQUEST_TIMEOUT,
        )
        with self.assertRaises(error):
            mock_client.perform_request("GET")

    @mock.patch("requests.Session.request", side_effect=requests.exceptions.Timeout)
    def test_check_access_handle_timeout_error(self, mock_request, mock_sleep):
        """
        Check whether the request back off properly for `check_access`
        method for 5 times in case of Timeout error.
        """
        mock_client = client.MixpanelClient(
            api_secret="mock_api_secret",
            api_domain="mock_api_domain",
            request_timeout=REQUEST_TIMEOUT,
        )
        with self.assertRaises(client.ReadTimeoutError):
            mock_client.check_access()

        # Verify that requests.Session.request is called 5 times
        self.assertEqual(mock_request.call_count, 5)

    @mock.patch("requests.Session.request")
    @mock.patch("tap_mixpanel.client.LOGGER.warning")
    def test_check_access_402_exception_handling(
        self, mock_logger, mock_request, mock_sleep
    ):
        """
        Test that `check_access` method does not throw 402 error and prints
        warning logger instead.
        """
        mock_request.return_value = MockResponse(402)
        mock_client = client.MixpanelClient(
            api_secret="mock_api_secret",
            api_domain="mock_api_domain",
            request_timeout=REQUEST_TIMEOUT,
        )

        mock_client.check_access()

        # Verify that for 402 error expected logger is printed.
        mock_logger.assert_called_with(
            "Mixpanel returned a 402 from the Engage API. Engage stream will be skipped."
        )


# Mock time.sleep to reduce the time
@mock.patch("time.sleep", return_value=None)
class TestMixpanelConnectionResetErrorHandling(unittest.TestCase):

    @mock.patch("requests.Session.request", side_effect=requests.models.ProtocolError)
    def test_check_access_handle_timeout_error(self, mock_request, mock_time):
        """
        Check whether the request backoffs properly for `check_access` method for 5 times in case of Timeout error.
        """
        mock_client = client.MixpanelClient(api_secret="mock_api_secret", api_domain="mock_api_domain", request_timeout=REQUEST_TIMEOUT)
        with self.assertRaises(requests.models.ProtocolError):
            mock_client.check_access()

        # Verify that requests.Session.request is called 5 times
        self.assertEqual(mock_request.call_count, 5)



import unittest
from unittest import mock
from parameterized import parameterized
from singer.catalog import Catalog
from tap_mixpanel.discover import discover
from tap_mixpanel.schema import get_schema, get_schemas
from tap_mixpanel.client import MixpanelPaymentRequiredError

@mock.patch("tap_mixpanel.schema.get_schema")
class TestDiscover(unittest.TestCase):
    """Test `discover` function."""

    def test_discover(self, mock_schema):
        """
        Test discover function generates a catalog object.
        """
        mock_schema.return_value = dict()
        client = mock.Mock()
        return_catalog = discover(client=client, properties_flag=False)

        # Verify discover function returns `Catalog` type object.
        self.assertIsInstance(return_catalog, Catalog)


class TestGetSchema(unittest.TestCase):
    """
    Test get_schema returns schema for dynamic and standard schemas.
    """

    engage_schema_response = {
        "status": "ok",
        "results": {
            "$last_seen": {"type": "datetime"},
            "country": {"type": "string"}
        }
    }

    export_schema_response = {
        "$last_seen": {},
        "country": {}
    }

    @parameterized.expand([
        ["engage", "engage/properties", engage_schema_response],
        ["export", "events/properties/top", export_schema_response],
    ])
    def test_dynamic_schema(self, stream_name, stream_path, schema_response):
        """
        Test that streams with dynamic schema function makes http call
        and writes fields as per the response.
        """
        client = mock.Mock()
        client.request.return_value = schema_response
        schema = get_schema(client=client, properties_flag=True, stream_name=stream_name)

        # Verify that request method is called with the path to get stream schema
        self.assertTrue(client.request.called)
        _, args = client.request.call_args
        self.assertEqual(stream_path, args["path"])

        # Verify that field without '$' is written as it is
        self.assertIn("country", schema["properties"])

        # Verify that field with '$' is written with 'mp_reserved_'
        self.assertIn("mp_reserved_last_seen", schema["properties"])

    def test_other_schema(self):
        """
        Test for standard stream function is not making http call to get schema.
        """
        client = mock.Mock()
        client.request.return_value = mock.Mock()
        schema = get_schema(client=client, properties_flag=True, stream_name="revenue")

        # Verify that request method is not called
        self.assertFalse(client.request.called)

        # Verify that schema type is dictionary
        self.assertEqual(type(schema), dict)

    @mock.patch("tap_mixpanel.schema.LOGGER.warning")
    def test_402_handling(self, mock_logger):
        """
        Test that while generating dynamic schema if 402 error occurs,
        tap does not throw the error and instead prints an warning logger.
        """
        client = mock.Mock()
        client.request.side_effect = MixpanelPaymentRequiredError

        schemas, field_metadata = get_schemas(client, True)

        # Verify that expected warning logger is called
        mock_logger.assert_any_call(
            "Mixpanel returned a 402 indicating the Engage endpoint and stream is unavailable. Skipping."
        )
        mock_logger.assert_any_call(
            "Mixpanel returned a 402 from the %s API so %s stream will be skipped.",
                "export",
                "export",
        )

        # Verify that dynamic schema stream is not written in catalog.
        self.assertNotIn("export", schemas)
        self.assertNotIn("engage", schemas)



import unittest

from parameterized import parameterized
from tap_mixpanel.transform import transform_record


class TestTransformRecord(unittest.TestCase):
    """
    Test case for transforming record.
    """

    test_dict_1 = {
        "$distinct_id": "1234",
        "$properties": {
            "$key1": "value1",
            "$key2": "value2"
        }
    }
    exp_dict_1 = {
        "distinct_id": "1234",
        "mp_reserved_key1": "value1",
        "mp_reserved_key2": "value2"
    }
    test_dict_2 = {
        "time": "1665052647",
        "properties": {
            "$key1": "value1",
            "$key2": "value2",
            "key3": "value3"
        }
    }
    exp_dict_2 = {
        "mp_reserved_key1": "value1",
        "mp_reserved_key2": "value2",
        "key3": "value3",
        "time": "2022-10-06T10:37:27.000000Z"
    }
    test_dict_3 = {
        "key1": "value1",
        "key2": "value2"
    }
    exp_dict_3 = {
        "key1": "value1",
        "key2": "value2",
        "parent_key1": "parent_value1"
    }
    test_dict_4 = {
        "$distinct_id": "1234",
    }
    exp_dict_4 = {
        "cohort_id": "5678",
        "distinct_id": "1234"
    }
    test_dict_5 = {
        "key1": "value1",
        "key2": "value2"
    }

    @parameterized.expand([
        ["engage", test_dict_1, exp_dict_1],
        ["export", test_dict_2, exp_dict_2],
        ["funnels", test_dict_3, exp_dict_3],
        ["cohort_members", test_dict_4, exp_dict_4],
        ["cohort", {**test_dict_5}, test_dict_5],
    ])
    def test_transform_record(self, stream, test_dict_1, expected_dict):
        """
        Test that `transform_record` function transforms records respected by streams.
        """

        parent_record = None
        project_timezone = "UTC"
        if stream == "funnels":
            parent_record = {"parent_key1": "parent_value1"}
        elif stream == "cohort_members":
            parent_record = {"id": "5678"}
        transformed_dict = transform_record(test_dict_1, stream, project_timezone, parent_record)

        # Verify that returned record is expected
        self.assertEqual(transformed_dict, expected_dict)



from singer.catalog import Catalog, CatalogEntry, Schema

from tap_mixpanel.schema import get_schemas
from tap_mixpanel.streams import STREAMS


def discover(client, properties_flag):
    """Run the discovery mode, prepare the catalog file and return catalog.

    Args:
        client (MixpanelClient): Client object to make http calls.
        properties_flag (str): Setting this argument to `true` ensures that new properties on
                               events and engage records are captured.

    Returns:
        singer.Catalog: Catalog object having schema and metadata of all the streams.
    """
    schemas, field_metadata = get_schemas(client, properties_flag)
    catalog = Catalog([])

    for stream_name, schema_dict in schemas.items():
        schema = Schema.from_dict(schema_dict)
        mdata = field_metadata[stream_name]

        catalog.streams.append(
            CatalogEntry(
                stream=stream_name,
                tap_stream_id=stream_name,
                key_properties=STREAMS[stream_name].key_properties,
                schema=schema,
                metadata=mdata,
            )
        )

    return catalog



import base64

import backoff
import jsonlines
import requests
import singer
from requests.exceptions import ConnectionError, Timeout
from requests.models import ProtocolError
from singer import metrics

LOGGER = singer.get_logger()

BACKOFF_MAX_TRIES_REQUEST = 7
REQUEST_TIMEOUT = 300


class ReadTimeoutError(Exception):
    """Custom error for request timeout."""


class Server5xxError(Exception):
    """Custom error class for all the 5xx error."""


class Server429Error(Exception):
    """Custom error class for rate limit exceeded."""


class MixpanelError(Exception):
    """Custom error class for all the Mixpanel errors."""


class MixpanelBadRequestError(MixpanelError):
    """Custom error class for bad request."""


class MixpanelUnauthorizedError(MixpanelError):
    """Custom error class for authorization."""


class MixpanelPaymentRequiredError(MixpanelError):
    """Custom error if API call require payment."""


class MixpanelNotFoundError(MixpanelError):
    """Custom error class for not found error."""


class MixpanelForbiddenError(MixpanelError):
    """Custom error class for forbidden error."""


class MixpanelInternalServiceError(Server5xxError):
    """Custom error class for internal server error."""


# Custom errors with respective messages mapped by error code.
ERROR_CODE_EXCEPTION_MAPPING = {
    400: {
        "raise_exception": MixpanelBadRequestError,
        "message": "A validation exception has occurred.",
    },
    401: {
        "raise_exception": MixpanelUnauthorizedError,
        "message": "Invalid authorization credentials.",
    },
    402: {
        "raise_exception": MixpanelPaymentRequiredError,
        "message": "Your current plan does not allow API calls. Payment is required to complete the operation.",
    },
    403: {
        "raise_exception": MixpanelForbiddenError,
        "message": "User does not have permission to access the resource.",
    },
    404: {
        "raise_exception": MixpanelNotFoundError,
        "message": "The resource you have specified cannot be found.",
    },
    429: {
        "raise_exception": Server429Error,
        "message": "The API rate limit for your organization/application pairing has been exceeded.",
    },
    500: {
        "raise_exception": MixpanelInternalServiceError,
        "message": "Server encountered an unexpected condition that prevented it from fulfilling the request.",
    },
}


def raise_for_error(response):
    """Retrieve the error code and the error message from the response
       and raises custom exceptions accordingly.

    Args:
        response (requests.Response): Response with error code.

    Raises:
        exc: Custom exception prepared according to status code.
    """
    LOGGER.error(
        "ERROR %s: %s, REASON: %s", response.status_code, response.text, response.reason
    )
    try:
        response_json = response.json()
    except Exception:
        response_json = {}
    error_code = response.status_code
    error_message = response_json.get(
        "error",
        response_json.get(
            "message",
            ERROR_CODE_EXCEPTION_MAPPING.get(error_code, {}).get(
                "message", "Unknown Error"
            ),
        ),
    )

    # If response text contains something unusual error of to_date then provide helper message of timezone mismatch
    # E.g error: to_date cannot be later than today
    if error_code == 400:
        if "to_date" in response.text:
            error_message += " Please validate the timezone with the MixPanel UI under project settings."
        else:
            error_message = f"{error_message}(Please verify your credentials.)"

    message = f"HTTP-error-code: {error_code}, Error: {error_message}"

    exc = ERROR_CODE_EXCEPTION_MAPPING.get(error_code, {}).get(
        "raise_exception", MixpanelError
    )
    raise exc(message) from None


class MixpanelClient:
    """
    The client class used for making REST calls to the Mixpanel API.
    """
    def __init__(self, api_secret, api_domain, request_timeout, user_agent=None):
        self.__api_secret = api_secret
        self.__api_domain = api_domain
        self.__request_timeout = request_timeout
        self.__user_agent = user_agent
        self.__session = requests.Session()
        self.__verified = False
        self.disable_engage_endpoint = False

    def __enter__(self):
        self.__verified = self.check_access()
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.__session.close()

    @backoff.on_exception(
        backoff.expo,
        (Server5xxError, Server429Error, ReadTimeoutError, ConnectionError, Timeout, ProtocolError),
        max_tries=5,
        factor=2,
    )
    def check_access(self):
        """Call rest API to verify user's credentials.

        Raises:
            Exception: Raises if response is not success.
            ReadTimeoutError: Raises if requests timeout.

        Returns:
            bool: Returns true if credentials are verified.
                  (else raises Exception)
        """
        if self.__api_secret is None:
            raise Exception("Error: Missing api_secret in tap config.json.")
        headers = {}
        # Endpoint: simple API call to return a single record (org settings) to test access
        url = f"https://{self.__api_domain}/api/2.0/engage"
        if self.__user_agent:
            headers["User-Agent"] = self.__user_agent
        headers["Accept"] = "application/json"
        headers[
            "Authorization"
        ] = f"Basic {str(base64.urlsafe_b64encode(self.__api_secret.encode('utf-8')), 'utf-8')}"

        try:
            response = self.__session.get(
                url=url,
                timeout=self.__request_timeout,  # Request timeout parameter
                headers=headers,
            )
        except requests.exceptions.Timeout as err:
            LOGGER.error("TIMEOUT ERROR: %s", str(err))
            raise ReadTimeoutError from None

        if response.status_code == 402:
            # 402 Payment Requirement does not indicate a permissions or authentication error
            self.disable_engage_endpoint = True
            LOGGER.warning(
                "Mixpanel returned a 402 from the Engage API. Engage stream will be skipped."
            )
            return True
        elif response.status_code != 200:
            LOGGER.error("Error status_code = %s", response.status_code)
            raise_for_error(response)
        return True

    @backoff.on_exception(
        backoff.expo,
        (Server5xxError, Server429Error, ReadTimeoutError, ConnectionError, Timeout, ProtocolError),
        max_tries=BACKOFF_MAX_TRIES_REQUEST,
        factor=3,
        logger=LOGGER,
    )
    def perform_request(
        self, method, url=None, params=None, json=None, stream=False, **kwargs
    ):
        """Call rest API and return the response in case of status code 200.

        Args:
            method (str): GET or POST method.
            url (str, optional): Complete url for the stream. Defaults to None.
            params (dict, optional): Query params. Defaults to None.
            json (dict, optional): JSON data (For POST request). Defaults to None.
            stream (bool, optional): If False, a response transfers indicating that
                                     the file should download immediately. If True, stream the file.
                                     Defaults to False.

        Raises:
            Server5xxError: Raises if status code > 500
            ReadTimeoutError: Raises if request timeouts.

        Returns:
            dict: With status code 200, returns JSON formatted response.
        """
        try:
            response = self.__session.request(
                method=method,
                url=url,
                params=params,
                json=json,
                stream=stream,
                timeout=self.__request_timeout,  # Request timeout parameter
                **kwargs,
            )

            if response.status_code > 500:
                raise Server5xxError()

            if response.status_code != 200:
                raise_for_error(response)
            return response
        except requests.exceptions.Timeout as err:
            LOGGER.error("TIMEOUT ERROR: %s", str(err))
            raise ReadTimeoutError(err) from None

    def request(self, method, url=None, path=None, params=None, json=None, **kwargs):
        """Request method to return JSON response of HTTP call.

        Args:
            method (str): GET or POST method.
            url (str, optional): Base URL. Defaults to None.
            path (str, optional): Path for the stream. Defaults to None.
            params (dict, optional): Query params. Defaults to None.
            json (dict, optional): JSON data (For POST requests). Defaults to None.

        Returns:
            dict: JSON object of response.
        """
        if not self.__verified:
            self.__verified = self.check_access()

        if url and path:
            url = f"{url}/{path}"
        elif path and not url:
            url = f"https://{self.__api_domain}/api/2.0/{path}"

        if "endpoint" in kwargs:
            endpoint = kwargs["endpoint"]
            del kwargs["endpoint"]
        else:
            endpoint = None

        if "headers" not in kwargs:
            kwargs["headers"] = {}

        kwargs["headers"]["Accept"] = "application/json"

        if self.__user_agent:
            kwargs["headers"]["User-Agent"] = self.__user_agent

        if method == "POST":
            kwargs["headers"]["Content-Type"] = "application/json"

        kwargs["headers"][
            "Authorization"
        ] = f"Basic {str(base64.urlsafe_b64encode(self.__api_secret.encode('utf-8')), 'utf-8')}"
        with metrics.http_request_timer(endpoint) as timer:
            response = self.perform_request(
                method=method, url=url, params=params, json=json, **kwargs
            )

            timer.tags[metrics.Tag.http_status_code] = response.status_code

        response_json = response.json()
        return response_json

    def request_export(
        self, method, url=None, path=None, params=None, json=None, **kwargs
    ):
        """Method to read jsonline from export stream response.

        Args:
            method (str): HTTP request method.
            url (str, optional): Base URL for the export endpoint. Defaults to None.
            path (str, optional): Path to the stream(export). Defaults to None.
            params (dict, optional): Request calls params. Defaults to None.
            json (dict, optional): JSON data (For POST request). Defaults to None.

        Yields:
            dict: Records of export stream.
        """
        if not self.__verified:
            self.__verified = self.check_access()

        if url and path:
            url = f"{url}/{path}"
        elif path and not url:
            url = f"https://{self.__api_domain}/api/2.0/{path}"

        if "endpoint" in kwargs:
            endpoint = kwargs["endpoint"]
            del kwargs["endpoint"]
        else:
            endpoint = "export"

        if "headers" not in kwargs:
            kwargs["headers"] = {}

        kwargs["headers"]["Accept"] = "application/json"

        if self.__user_agent:
            kwargs["headers"]["User-Agent"] = self.__user_agent

        if method == "POST":
            kwargs["headers"]["Content-Type"] = "application/json"

        kwargs["headers"][
            "Authorization"
        ] = f"Basic {str(base64.urlsafe_b64encode(self.__api_secret.encode('utf-8')), 'utf-8')}"
        with metrics.http_request_timer(endpoint) as timer:
            response = self.perform_request(
                method=method, url=url, params=params, json=json, stream=True, **kwargs
            )
            timer.tags[metrics.Tag.http_status_code] = response.status_code

            # 'export' endpoint returns jsonl results;
            #  Other endpoints return json with array of results
            #  jsonlines reference: https://jsonlines.readthedocs.io/en/latest/
            reader = jsonlines.Reader(response.iter_lines())
            yield from reader.iter(allow_none=True, skip_empty=True)



import pytz
import unittest

from datetime import datetime
from tap_mixpanel.transform import transform_event_times

UTC = pytz.utc


class TestTransformEventTimes(unittest.TestCase):
    """
    Test that `transform_event_times` function formats,
    the Eastern and UTC formatted dates to ISO datetime.
    """

    def test_utc_now(self):
        """
        Testcase for the UTC timezone is converted to the given format.
        """

        input_time = datetime.utcnow()

        record = {"time": input_time.timestamp()}
        project_timezone = "UTC"

        actual = transform_event_times(record, project_timezone)
        expected = {
            "time": input_time.astimezone(UTC).strftime("%04Y-%m-%dT%H:%M:%S.000000Z")
        }

        # Verify that record uis converted as expected.
        self.assertEqual(expected, actual)

    def test_eastern_time(self):
        """
        Testcase for the eastern timezone is converted to given formate.
        """

        project_timezone = "US/Eastern"
        EASTERN = pytz.timezone(project_timezone)
        # This gives us 2021-08-12T11:00:00-4:00
        input_time = EASTERN.localize(datetime(2021, 8, 12, 11, 0, 0))

        record = {"time": input_time.timestamp()}
        actual = transform_event_times(record, project_timezone)

        expected = {
            "time": input_time.astimezone(UTC).strftime("%04Y-%m-%dT%H:%M:%S.000000Z")
        }

        # Verify that record uis converted as expected.
        self.assertEqual(expected, actual)



import datetime

import pytz
import singer
from singer import Transformer
from singer.utils import strftime

LOGGER = singer.get_logger()


def denest_properties(record, properties_node):
    """De-nest properties for engage and export endpoints. Write fields to
    first level from `properties_node`.

    Args:
        record (dict): Record to update.
        properties_node (str): Nested object whose fields will be written at 1st level.

    Returns:
        dict: Updated record
    """
    new_record = record
    properties = record.get(properties_node)
    if properties:
        for key, val in record[properties_node].items():
            if key[0:1] == "$":
                new_key = f"mp_reserved_{key[1:]}"
                # Change this to regex
            else:
                new_key = key
            new_record[new_key] = val
        new_record.pop(properties_node, None)
    return new_record


# Reference: https://help.mixpanel.com/hc/en-us/articles/115004547203-Manage-Timezones-for-Projects-in-Mixpanel#exporting-data-from-mixpanel
def transform_event_times(record, project_timezone):
    """Time conversion from $time integer using project_timezone.

    Args:
        record (dict): Record to be transform.
        project_timezone (str): Time zone in which integer date times are stored.

    Returns:
        dict: Updated record.
    """
    new_record = record
    timezone = pytz.timezone(project_timezone)

    # Create beginning_datetime: beginning of epoch time in project timezone
    naive_time = datetime.time(0, 0)
    date = datetime.date(1970, 1, 1)
    naive_datetime = datetime.datetime.combine(date, naive_time)
    # Move 1970-01-01T00:00:00Z to Mixpanel project timezone.
    # For example, if Mixpanel timezone is in Eastern Time (UTC-5:00) then it calculates 1969-12-31T19:00:00-5:00.
    beginning_datetime = pytz.utc.localize(naive_datetime).astimezone(timezone)

    # Get integer time
    time_int = int(record.get("time"))

    # Create new_time_utc by adding seconds to beginning_datetime, normalizing,
    #   and converting to string
    add_seconds = datetime.timedelta(seconds=time_int)
    new_time = beginning_datetime + add_seconds

    # 'normalize' accounts for daylight savings time
    new_time_utc_str = strftime(timezone.normalize(new_time).astimezone(pytz.utc))
    new_record["time"] = new_time_utc_str

    return new_record


def transform_datetime(this_dttm):
    """Transform date_time string TO DATETIME object.

    Args:
        this_dttm (str): Formatted date-time string

    Returns:
        datetime: Datetime object passed string.
    """
    with Transformer() as transformer:
        new_dttm = transformer._transform_datetime(this_dttm)
    return new_dttm


def transform_engage(record):
    """Remove leading $ from engage $distinct_id.

    Args:
        record (dict): record to be update.

    Returns:
        dict: New updated record.
    """
    new_record = record
    distinct_id = record.get("$distinct_id")
    new_record["distinct_id"] = distinct_id
    new_record.pop("$distinct_id", None)
    return new_record


def transform_funnels(record, parent_record):
    """Funnels: Combine parent record with each date record

    Args:
        record (dict): Record to be transform.
        parent_record (dict): Parent record.

    Returns:
        dict: Updated record.
    """
    record.update(parent_record)
    return record


def transform_cohort_members(record, parent_record):
    """Cohort Members: provide all distinct_id's for each cohort_id.

    Args:
        record (dict): Record to be transform.
        parent_record (dict): Parent stream  record.

    Returns:
        dict: Record with id fields.
    """
    cohort_id = parent_record.get("id")
    distinct_id = record.get("$distinct_id")
    new_record = {}
    new_record["distinct_id"] = distinct_id
    new_record["cohort_id"] = cohort_id
    return new_record


# Run other transforms, as needed: denest_list_nodes, transform_conversation_parts
def transform_record(record, stream_name, project_timezone, parent_record=None):
    """Transform record and add fields at first level as required by stream.

    Args:
        record (dict): Record to be transform.
        stream_name (str): Stream that record belongs to.
        project_timezone (str): Time zone in which integer date times are stored.
        parent_record (dict, optional): Parent stream record if current stream is child.
                                        Defaults to None.

    Returns:
        dict: Transformed record.
    """
    if stream_name == "engage":
        trans_json = transform_engage(record)
        new_record = denest_properties(trans_json, "$properties")
    elif stream_name == "export":
        denested_json = denest_properties(record, "properties")
        new_record = transform_event_times(denested_json, project_timezone)
    elif stream_name == "funnels":
        new_record = transform_funnels(record, parent_record)
    elif stream_name == "cohort_members":
        new_record = transform_cohort_members(record, parent_record)
    else:
        new_record = record

    return new_record



import unittest
from unittest import mock

from tap_mixpanel import client


class MockResponse:
    """Mocked standard HTTPResponse to test error handling."""
    status_code = 200

    def iter_lines(self):
        """Mock generator(list) to return in response."""
        return [
            b'{"event": "Page View", "properties": {"time": 1583044147, "distinct_id": "test_id_1"}}',
            b'{"event": "Page View", "properties": {"time": 1583225657, "distinct_id": "test_id_2"}}'
        ]

    def json(self):
        """JSON formatted response"""
        return {}


class TestClientMethods(unittest.TestCase):
    """
    Test case to validate working of client methods.
    """

    @mock.patch("tap_mixpanel.client.MixpanelClient.check_access")
    @mock.patch("tap_mixpanel.client.MixpanelClient.perform_request")
    def test_request_with_url(self, mock_perform_request, mock_check_access):
        """
        Test `request` method for success response with URL and Path.
        """

        mock_check_access.return_value = True
        mock_perform_request.return_value = MockResponse()
        mock_client = client.MixpanelClient(
            api_secret="mock_api_secret",
            api_domain="mock_api_domain",
            request_timeout=300,
            user_agent="USER_AGENT"
        )

        response = mock_client.request(method="GET", url="https://sample_url", path="sample_path",
                                       endpoint={})

        # Verify that check_access is called.
        self.assertTrue(mock_check_access.called)

        # Verify that returned response is expected.
        self.assertEqual(response, {})

    @mock.patch("tap_mixpanel.client.MixpanelClient.check_access")
    @mock.patch("tap_mixpanel.client.MixpanelClient.perform_request")
    def test_request_without_url(self, mock_perform_request, mock_check_access):
        """
        Test `request` method for success response without URL.
        """

        mock_check_access.return_value = True
        mock_perform_request.return_value = MockResponse()
        mock_client = client.MixpanelClient(
            api_secret="mock_api_secret",
            api_domain="mock_api_domain",
            request_timeout=300,
            user_agent="USER_AGENT"
        )

        response = mock_client.request(method="GET", path="sample_path", endpoint={})

        # Verify that check_access is called.
        self.assertTrue(mock_check_access.called)

        # Verify that returned response is expected.
        self.assertEqual(response, {})

    @mock.patch("tap_mixpanel.client.MixpanelClient.check_access")
    @mock.patch("tap_mixpanel.client.MixpanelClient.perform_request")
    def test_request_export_with_url(self, mock_perform_request, mock_check_access):
        """
        Test `request_export` with passing base URL for success response.
        """

        mock_perform_request.return_value = MockResponse()
        mock_client = client.MixpanelClient(
            api_secret="mock_api_secret",
            api_domain="mock_api_domain",
            request_timeout=300,
            user_agent="USER_AGENT"
        )

        response = list(mock_client.request_export(
            method="POST", url="https://sample_url", path="sample_path", endpoint={}))

        expected_data = [
            {"event": "Page View",
             "properties": {
                 "time": 1583044147,
                 "distinct_id": "test_id_1"}},
            {"event": "Page View",
             "properties": {
                 "time": 1583225657,
                 "distinct_id": "test_id_2"}}
        ]

        # Verify that check_access is called.
        self.assertTrue(mock_check_access.called)

        # Verify that returned response is expected.
        self.assertEqual(list(response), expected_data)

    @mock.patch("tap_mixpanel.client.MixpanelClient.check_access")
    @mock.patch("tap_mixpanel.client.MixpanelClient.perform_request")
    def test_request_export_without_url(self, mock_perform_request, mock_check_access):
        """
        Test `request_export` without passing base URL for success response.
        """

        mock_perform_request.return_value = MockResponse()
        mock_client = client.MixpanelClient(
            api_secret="mock_api_secret",
            api_domain="mock_api_domain",
            request_timeout=300,
            user_agent="USER_AGENT"
        )

        response = list(mock_client.request_export(method="POST", path="sample_path", endpoint={}))

        expected_data = [
            {"event": "Page View",
             "properties": {
                 "time": 1583044147,
                 "distinct_id": "test_id_1"}},
            {"event": "Page View",
             "properties": {
                 "time": 1583225657,
                 "distinct_id": "test_id_2"}}
        ]

        # Verify that check_access method is called.
        self.assertTrue(mock_check_access.called)

        # Verify that returned response is expected.
        self.assertEqual(list(response), expected_data)



from tap_tester import connections, runner, LOGGER
from base import TestMixPanelBase


class MixPanelStartDateTest(TestMixPanelBase):
    start_date_1 = ""
    start_date_2 = ""

    @staticmethod
    def name():
        return "tap_tester_mixpanel_start_date_test"

    def start_date_test_run(self):
        """Instantiate start date according to the desired data set and run the test."""

        self.start_date_1 = self.get_properties().get("start_date")
        self.start_date_2 = self.timedelta_formatted(self.start_date_1, days=15)

        self.start_date = self.start_date_1

        expected_streams = self.expected_streams()

        ##########################################################################
        # First Sync
        ##########################################################################

        # Instantiate connection
        conn_id_1 = connections.ensure_connection(self)

        # Run check mode
        found_catalogs_1 = self.run_and_verify_check_mode(conn_id_1)

        # Table and field selection
        test_catalogs_1_all_fields = [
            catalog
            for catalog in found_catalogs_1
            if catalog.get("tap_stream_id") in expected_streams
        ]
        self.perform_and_verify_table_and_field_selection(
            conn_id_1, test_catalogs_1_all_fields, select_all_fields=True
        )

        # Run initial sync
        record_count_by_stream_1 = self.run_and_verify_sync(conn_id_1)
        synced_records_1 = runner.get_records_from_target_output()

        ##########################################################################
        # Update START DATE Between Syncs
        ##########################################################################

        LOGGER.info(
            f"REPLICATION START DATE CHANGE: {self.start_date} ===>>> {self.start_date_2} "
        )
        self.start_date = self.start_date_2

        ##########################################################################
        # Second Sync
        ##########################################################################

        # Create a new connection with the new start_date
        conn_id_2 = connections.ensure_connection(self, original_properties=False)

        # Run check mode
        found_catalogs_2 = self.run_and_verify_check_mode(conn_id_2)

        # Table and field selection
        test_catalogs_2_all_fields = [
            catalog
            for catalog in found_catalogs_2
            if catalog.get("tap_stream_id") in expected_streams
        ]
        self.perform_and_verify_table_and_field_selection(
            conn_id_2, test_catalogs_2_all_fields, select_all_fields=True
        )

        # Run sync
        record_count_by_stream_2 = self.run_and_verify_sync(conn_id_2)
        synced_records_2 = runner.get_records_from_target_output()

        for stream in expected_streams:
            with self.subTest(stream=stream):

                # Expected values
                expected_primary_keys = self.expected_pks()[stream]
                expected_metadata = self.expected_metadata()[stream]
                expected_start_date_1 = self.timedelta_formatted(self.start_date_1)
                expected_start_date_2 = self.timedelta_formatted(self.start_date_2)

                # Collect information for assertions from syncs 1 & 2 base on expected values
                record_count_sync_1 = record_count_by_stream_1.get(stream, 0)
                record_count_sync_2 = record_count_by_stream_2.get(stream, 0)

                primary_keys_list_1 = [
                    tuple(
                        message.get("data").get(expected_pk)
                        for expected_pk in expected_primary_keys
                    )
                    for message in synced_records_1.get(stream, {}).get("messages", [])
                    if message.get("action") == "upsert"
                ]
                primary_keys_list_2 = [
                    tuple(
                        message.get("data").get(expected_pk)
                        for expected_pk in expected_primary_keys
                    )
                    for message in synced_records_2.get(stream, {}).get("messages", [])
                    if message.get("action") == "upsert"
                ]

                primary_keys_sync_1 = set(primary_keys_list_1)
                primary_keys_sync_2 = set(primary_keys_list_2)

                if expected_metadata.get(self.OBEYS_START_DATE):

                    # Collect information specific to incremental streams from syncs 1 & 2
                    expected_replication_key = next(
                        iter(self.expected_replication_keys().get(stream, []))
                    )
                    replication_dates_1 = [
                        row.get("data").get(expected_replication_key)
                        for row in synced_records_1.get(stream, {"messages": []}).get(
                            "messages", []
                        )
                        if row.get("data")
                    ]
                    replication_dates_2 = [
                        row.get("data").get(expected_replication_key)
                        for row in synced_records_2.get(stream, {"messages": []}).get(
                            "messages", []
                        )
                        if row.get("data")
                    ]

                    # Verify replication key is greater or equal to start_date for sync 1
                    for replication_date in replication_dates_1:
                        self.assertGreaterEqual(
                            self.parse_date(replication_date),
                            self.parse_date(expected_start_date_1),
                            msg="Report pertains to a date prior to our start date.\n"
                            + f"Sync start_date: {expected_start_date_1}\n"
                            + f"Record date: {replication_date} ",
                        )

                    # Verify replication key is greater or equal to start_date for sync 2
                    for replication_date in replication_dates_2:
                        self.assertGreaterEqual(
                            self.parse_date(replication_date),
                            self.parse_date(expected_start_date_2),
                            msg="Report pertains to a date prior to our start date.\n"
                            + f"Sync start_date: {expected_start_date_2}\n"
                            + f"Record date: {replication_date} ",
                        )

                    # Verify the number of records replicated in sync 1 is greater than the number
                    # of records replicated in sync 2
                    self.assertGreater(record_count_sync_1, record_count_sync_2)

                    # Verify the records replicated in sync 2 were also replicated in sync 1
                    self.assertTrue(primary_keys_sync_2.issubset(primary_keys_sync_1))

                else:

                    # Verify that the 2nd sync with a later start date replicates the same number of
                    # records as the 1st sync.
                    self.assertEqual(record_count_sync_2, record_count_sync_1)

                    # Verify by primary key the same records are replicated in the 1st and 2nd syncs
                    self.assertSetEqual(primary_keys_sync_1, primary_keys_sync_2)

    def test_run(self):
        # Start date test for standard server
        self.eu_residency = False
        self.start_date_test_run()



import singer

from tap_mixpanel.streams import STREAMS

LOGGER = singer.get_logger()


def get_streams_to_sync(selected_streams):
    """Get lists of streams to call the sync method.

    For children, ensure that dependent parent_stream is included even
    if it is not selected.
    """
    streams_to_sync = []

    # Loop thru all selected streams
    for stream_name in selected_streams:
        stream_obj = STREAMS[stream_name]
        # If the stream has a parent_stream, then it is a child stream
        parent_stream = hasattr(stream_obj, "parent") and stream_obj.parent

        # Append selected parent streams
        if not parent_stream:
            streams_to_sync.append(stream_name)
        else:
            # Append un-selected parent streams of selected children
            if (
                parent_stream not in selected_streams
                and parent_stream not in streams_to_sync
            ):
                streams_to_sync.append(parent_stream)

    return streams_to_sync


def write_schemas_recursive(stream_id, catalog, selected_streams):
    """Write the schemas for the selected parent and it's all child."""
    # Passing None as client as we are only initializing object to write schema
    stream_obj = STREAMS[stream_id](None)

    if stream_id in selected_streams:
        stream_obj.write_schema(catalog, stream_id)

    # Write schema for selected child
    if stream_obj.child:
        write_schemas_recursive(stream_obj.child, catalog, selected_streams)


def update_currently_syncing(state, stream_name):
    """Currently syncing sets the stream currently being delivered in the
    state.

    If the integration is interrupted, this state property is used to identify
     the starting point to continue from.
    Reference: https://github.com/singer-io/singer-python/blob/master/singer/bookmarks.py#L41-L46
    """
    if (stream_name is None) and ("currently_syncing" in state):
        del state["currently_syncing"]
    else:
        singer.set_currently_syncing(state, stream_name)
    singer.write_state(state)


def sync(client, config, catalog, state, start_date):
    """
    Get selected_streams from catalog, based on state last_stream
    last_stream = Previous currently synced stream, if the load was interrupted
    """
    last_stream = singer.get_currently_syncing(state)
    LOGGER.info("last/currently syncing stream: %s", last_stream)
    selected_streams = []
    for stream in catalog.get_selected_streams(state):
        selected_streams.append(stream.stream)
    streams_to_sync = get_streams_to_sync(selected_streams)
    LOGGER.info("selected_streams: %s", selected_streams)
    LOGGER.info("streams_to_sync: %s", streams_to_sync)

    if not selected_streams:
        return

    # Loop through selected_streams
    for stream_name in streams_to_sync:
        stream_obj = STREAMS[stream_name](client)

        update_currently_syncing(state, stream_name)

        # Write schema of only selected streams in parent-child stream
        write_schemas_recursive(stream_name, catalog, selected_streams)

        LOGGER.info("START Syncing: %s", stream_name)
        endpoint_total = stream_obj.sync(
            catalog=catalog,
            state=state,
            config=config,
            start_date=start_date,
            selected_streams=selected_streams,
        )

        update_currently_syncing(state, None)
        LOGGER.info(
            "FINISHED Syncing: %s, Total endpoint records: %s",
            stream_name,
            endpoint_total,
        )



import re
from tap_tester import menagerie, connections, LOGGER

from base import TestMixPanelBase

class MixPanelDiscoverTest(TestMixPanelBase):
    """
        Testing that discovery creates the appropriate catalog with valid metadata.
        • Verify number of actual streams discovered match expected
        • Verify the stream names discovered were what we expect
        • Verify stream names follow naming convention
          streams should only have lowercase alphas and underscores
        • Verify there is only 1 top level breadcrumb
        • Verify replication key(s)
        • Verify primary key(s)
        • Verify that if there is a replication key we are doing INCREMENTAL otherwise FULL
        • Verify the actual replication matches our expected replication method
        • Verify that primary, replication and foreign keys
          are given the inclusion of automatic.
        • Verify that all other fields have inclusion of available metadata.
    """

    @staticmethod
    def name():
        return "tap_tester_mixpanel_discover_test"

    def discovery_test_run(self):

        region = "EU" if self.eu_residency else "Standard"
        LOGGER.info(f"Testing against {region} account.")

        self.assertion_logging_enabled = True

        streams_to_test = self.expected_streams()

        conn_id = connections.ensure_connection(self, payload_hook=None)

        # Verify that there are catalogs found
        found_catalogs = self.run_and_verify_check_mode(conn_id)

        # Verify stream names follow naming convention
        # streams should only have lowercase alphas and underscores
        found_catalog_names = {c["tap_stream_id"] for c in found_catalogs}
        self.assertTrue(
            all([re.fullmatch(r"[a-z_]+", name) for name in found_catalog_names]),
            logging="asserting all streams defined in catalog follow the naming convention '[a-z_]+'",
        )

        for stream in streams_to_test:
            with self.subTest(stream=stream):

                # Verify the catalog is found for a given stream
                catalog = next(
                    iter(
                        [
                            catalog
                            for catalog in found_catalogs
                            if catalog["stream_name"] == stream
                        ]
                    )
                )
                self.assertIsNotNone(
                    catalog, logging="Asserting entry is present in catalog"
                )

                # Collecting expected values
                expected_primary_keys = self.expected_pks()[stream]
                expected_replication_keys = self.expected_replication_keys()[stream]
                expected_automatic_fields = self.expected_automatic_fields().get(stream)
                expected_replication_method = self.expected_replication_method()[stream]

                # Collecting actual values...
                schema_and_metadata = menagerie.get_annotated_schema(
                    conn_id, catalog["stream_id"]
                )
                metadata = schema_and_metadata["metadata"]
                stream_properties = [
                    item for item in metadata if item.get("breadcrumb") == []
                ]
                actual_primary_keys = set(
                    stream_properties[0]
                    .get("metadata", {self.PRIMARY_KEYS: []})
                    .get(self.PRIMARY_KEYS, [])
                )
                actual_replication_keys = set(
                    stream_properties[0]
                    .get("metadata", {self.REPLICATION_KEYS: []})
                    .get(self.REPLICATION_KEYS, [])
                )
                actual_replication_method = (
                    stream_properties[0]
                    .get("metadata", {self.REPLICATION_METHOD: None})
                    .get(self.REPLICATION_METHOD)
                )
                actual_automatic_fields = {
                    item.get("breadcrumb", ["properties", None])[1]
                    for item in metadata
                    if item.get("metadata").get("inclusion") == "automatic"
                }

                actual_fields = []
                for md_entry in metadata:
                    if md_entry["breadcrumb"] != []:
                        actual_fields.append(md_entry["breadcrumb"][1])

                ##########################################################################
                # Metadata assertions
                ##########################################################################

                # Verify there is only 1 top level breadcrumb in metadata
                self.assertEqual(
                    len(stream_properties),
                    1,
                    logging="Asserting there is only 1 top level breadcrumb in metadata",
                )

                # Verify there is no duplicate metadata entries
                self.assertEqual(
                    len(actual_fields),
                    len(set(actual_fields)),
                    msg="Duplicates in the fields retrieved",
                )

                # Verify that if there is a replication key we are doing INCREMENTAL otherwise FULL
                if actual_replication_keys:
                    self.assertEqual(
                        actual_replication_method,
                        self.INCREMENTAL,
                        logging=f"Asserting replication method is {self.INCREMENTAL} when replication keys are defined",
                    )
                else:
                    self.assertEqual(
                        actual_replication_method,
                        self.FULL_TABLE,
                        logging=f"Asserting replication method is {self.FULL_TABLE} when replication keys are not defined",
                    )

                # Verify the actual replication matches our expected replication method
                self.assertEqual(
                    expected_replication_method,
                    actual_replication_method,
                    logging=f"Asserting replication method is {expected_replication_method}",
                )

                # Verify replication key(s)
                self.assertEqual(
                    expected_replication_keys,
                    actual_replication_keys,
                    logging=f"asserting replication keys are {expected_replication_keys}",
                )

                # Verify primary key(s) match expectations
                self.assertSetEqual(
                    expected_primary_keys,
                    actual_primary_keys,
                    logging=f"asserting primary keys are {expected_primary_keys}",
                )

                # Verify that primary keys and replication keys
                # are given the inclusion of automatic in metadata.
                self.assertSetEqual(
                    expected_automatic_fields,
                    actual_automatic_fields,
                    logging=f"asserting primary and replication keys {expected_automatic_fields} are automatic",
                )

                # Verify that all other fields have inclusion of available.
                # This assumes there are no unsupported fields for SaaS sources
                self.assertTrue(
                    all(
                        {
                            item.get("metadata").get("inclusion") == "available"
                         for item in metadata
                         if item.get("breadcrumb", []) != []
                         and item.get("breadcrumb", ["properties", None])[1]
                            not in actual_automatic_fields
                        }
                    ),
                    logging="Asserting non-key-property fields are available for field selection",
                )

    def test_standard_discovery(self):
        """Discovery test for standard server."""
        self.eu_residency = False
        self.discovery_test_run()

    def test_eu_discovery(self):
        """Discovery test for EU residency server."""
        self.eu_residency = True
        self.discovery_test_run()



from datetime import datetime
import unittest
import pytz
from parameterized import parameterized
from singer.utils import strptime_to_utc
from tap_mixpanel.streams import Export

NOW_TIME = datetime(year=2022, month=10, day=10).replace(tzinfo=pytz.UTC)


class TestAttributionWindow(unittest.TestCase):
    """
    Test that for attribution window > current_date - (start_date/bookmark),
    Sync start from start_date.
    """

    @parameterized.expand([
        ["start_date_before_attribution_window", "2022-09-24T00:00:00Z", "2022-10-05T00:00:00Z"],
        ["from_the_start_date", "2022-10-05T00:00:00Z", "2022-10-05T00:00:00Z"],
        ["attribution_window_before_start_date", "2022-10-07T00:00:00Z", "2022-10-05T00:00:00Z"],
    ])
    def test_start_window_value(self, test_name, last_datetime, expected_start_window):
        """
        Test given scenarios:
            - If start_date or bookmark is < attribution window, start_window will be as it is.
            - Other wise, start window will be start_date
        """

        stream = Export(None)
        date_window_size = 30
        attribution_window = 10

        start_window, _, _ = stream.define_bookmark_filters(days_interval=date_window_size,
                                                            last_datetime=last_datetime,
                                                            now_datetime=NOW_TIME,
                                                            attribution_window=attribution_window,
                                                            start_date="2022-10-05T00:00:00Z")

        # Verify that start window is expected
        self.assertEqual(start_window, strptime_to_utc(expected_start_window))



import unittest
from unittest import mock

from tap_mixpanel.__init__ import main
from tap_mixpanel.client import MixpanelClient
from tap_mixpanel.streams import Export, Revenue

EU_CONFIG = {
    "api_secret": "dummy_secret",
    "date_window_size": "30",
    "attribution_window": "5",
    "project_timezone": "Europe/Amsterdam",
    "select_properties_by_default": "true",
    "start_date": "2020-02-01T00:00:00Z",
    "user_agent": "tap-mixpanel <api_user_email@your_company.com>",
    "eu_residency": True,
    "end_date": "2020-03-02T00:00:00Z",
}
STANDARD_CONFIG = {
    "api_secret": "dummy_secret",
    "date_window_size": "30",
    "attribution_window": "5",
    "project_timezone": "Europe/Amsterdam",
    "select_properties_by_default": "true",
    "start_date": "2020-02-01T00:00:00Z",
    "user_agent": "tap-mixpanel <api_user_email@your_company.com>",
    "eu_residency": False,
    "end_date": "2020-03-02T00:00:00Z",
}


class MockStream:
    """Mock stream object class."""

    def __init__(self, stream):
        self.stream = stream


class MockCatalog:
    """Mock catalog object class."""

    def __init__(self, name):
        self.name = name

    def get_selected_streams(self, state):
        """Returns the list of selected stream objects."""
        return [MockStream(self.name)]


class MockParseArgs:
    """Mock args object class."""

    def __init__(self, state, discover, config):
        self.state = state
        self.discover = discover
        self.config = config


class MockResponse:
    """Mocked standard HTTPResponse to test error handling."""

    def __init__(
        self, resp, status_code, content=[""], headers=None, raise_error=False, text={}
    ):
        self.json_data = resp
        self.status_code = status_code
        self.content = content
        self.headers = headers
        self.raise_error = raise_error
        self.text = text
        self.reason = "error"

    def json(self):
        """Returns a JSON object of the result."""
        return self.text


class TestMixpanelSupportEuEndpoints(unittest.TestCase):
    """
    Test that europe domain support is working.
    """

    @mock.patch("tap_mixpanel.client.MixpanelClient.request")
    @mock.patch("tap_mixpanel.streams.MixPanel.write_bookmark")
    @mock.patch("tap_mixpanel.streams.MixPanel.write_schema")
    def test_support_eu_endpoints_except_export(
        self, mock_write_schema, mock_write_bookmark, mock_request
    ):
        """
        Test case for the streams other than export stream that,
        For eu_residency europe domain base url is called.
        And for eu_residency 'false' in the config, default domain URL is called.
        """
        mock_request.return_value = {}
        mock_write_schema.return_value = ""
        mock_write_bookmark.return_value = ""

        state = {}
        catalog = MockCatalog("revenue")

        client = MixpanelClient("", "", "")
        revenue_obj = Revenue(client)
        revenue_obj.sync(
            catalog=catalog,
            state=state,
            config=EU_CONFIG,
            start_date="2020-02-01T00:00:00Z",
            selected_streams=["revenue"],
        )

        # Verify that with EU config, base url has eu-domain.
        mock_request.assert_called_with(
            method="GET",
            url="https://eu.mixpanel.com/api/2.0",
            path="engage/revenue",
            params="unit=day&from_date=2020-02-01&to_date=2020-03-02",
            endpoint="revenue",
        )

        revenue_obj = Revenue(client)
        revenue_obj.sync(
            catalog=catalog,
            state=state,
            config=STANDARD_CONFIG,
            start_date="2020-02-01T00:00:00Z",
            selected_streams=["revenue"],
        )

        # Verify that with standard config, base URL has default domain.
        mock_request.assert_called_with(
            method="GET",
            url="https://mixpanel.com/api/2.0",
            path="engage/revenue",
            params="unit=day&from_date=2020-02-01&to_date=2020-03-02",
            endpoint="revenue",
        )

    @mock.patch("tap_mixpanel.client.MixpanelClient.request_export")
    @mock.patch("tap_mixpanel.streams.MixPanel.write_bookmark")
    @mock.patch("tap_mixpanel.streams.MixPanel.write_schema")
    def test_support_export_eu_endpoint(
        self, mock_write_schema, mock_write_bookmark, mock_request_export
    ):
        """
        Test case for the export stream (as it has different base url) that,
        For eu_residency europe domain base url is called.
        And for eu_residency 'false' in the config, default domain URL is called.
        """
        mock_request_export.return_value = {}
        mock_write_schema.return_value = ""
        mock_write_bookmark.return_value = ""

        state = {}
        catalog = MockCatalog("export")

        client = MixpanelClient("", "", "")
        export_obj = Export(client)
        export_obj.sync(
            catalog=catalog,
            state=state,
            config=EU_CONFIG,
            start_date="2020-02-01T00:00:00Z",
            selected_streams=["exports"],
        )

        # Verify that with EU config, base url has eu-domain.
        mock_request_export.assert_called_with(
            method="GET",
            url="https://data-eu.mixpanel.com/api/2.0",
            path="export",
            params="from_date=2020-02-01&to_date=2020-03-02",
            endpoint="export",
        )

        export_obj = Export(client)
        export_obj.sync(
            catalog=catalog,
            state=state,
            config=STANDARD_CONFIG,
            start_date="2020-02-01T00:00:00Z",
            selected_streams=["exports"],
        )

        # Verify that with standard config, base URL has default domain.
        mock_request_export.assert_called_with(
            method="GET",
            url="https://data.mixpanel.com/api/2.0",
            path="export",
            params="from_date=2020-02-01&to_date=2020-03-02",
            endpoint="export",
        )

    @mock.patch("requests.Session.request")
    @mock.patch("singer.utils.parse_args")
    @mock.patch("tap_mixpanel.__init__.do_discover")
    def test_support_eu_endpoint_in_discover(
        self, mock_discover, mock_parse_args, mock_request
    ):
        """
        Test case for the discover mode,
        For eu_residency europe domain base url is called.
        And for eu_residency 'false' in the config, default domain URL is called.
        """

        mock_request.return_value = MockResponse("", status_code=200)
        mock_discover.return_value = ""
        mock_parse_args.return_value = MockParseArgs(
            state={}, discover=True, config=EU_CONFIG
        )
        header = {
            "User-Agent": "tap-mixpanel <api_user_email@your_company.com>",
            "Accept": "application/json",
            "Authorization": "Basic ZHVtbXlfc2VjcmV0",
        }

        main()

        # Verify that with EU config, base url has eu-domain.
        mock_request.assert_called_with(
            "GET",
            "https://eu.mixpanel.com/api/2.0/engage",
            allow_redirects=True,
            headers=header,
            timeout=300,
        )

        mock_parse_args.return_value = MockParseArgs(
            state={}, discover=True, config=STANDARD_CONFIG
        )
        main()

        # Verify that with standard config, base URL has default domain.
        mock_request.assert_called_with(
            "GET",
            "https://mixpanel.com/api/2.0/engage",
            allow_redirects=True,
            headers=header,
            timeout=300,
        )



import json
import os

import singer
from singer import metadata

from tap_mixpanel.client import MixpanelPaymentRequiredError
from tap_mixpanel.streams import STREAMS

LOGGER = singer.get_logger()


# Reference:
# https://github.com/singer-io/getting-started/blob/master/docs/DISCOVERY_MODE.md#Metadata


def get_abs_path(path):
    """Get the absolute path for the schema files.

    Args:
        path (str): Path from current folder to schema file.

    Returns:
        str: Full path to schema file.
    """
    return os.path.join(os.path.dirname(os.path.realpath(__file__)), path)


def get_schema(client, properties_flag, stream_name):
    """Creates schema for a stream by loading schema file and appending dynamic
    fields schema if necessary.

    Args:
        client (MixpanelClient): Client to make http calls.
        properties_flag (str): Setting this argument to `true` ensures that new properties on
                               events and engage records are captured.
        stream_name (str): Name of stream whose schema is to create.

    Returns:
        dict: Returns schema of the stream.
    """
    schema_path = get_abs_path(f"schemas/{stream_name}.json")

    with open(schema_path, encoding="utf-8") as file:
        schema = json.load(file)

    # Set whether to allow additional properties for engage and export endpoints
    # Event and Engage properties are dynamic and depend on the properties provided on upload,
    #   when the Event or Engage (user/person) was created.
    # Depending on the tap config parameter select_properties_by_default,
    #   the json schema should allow additional properties (additionalProperties = true).
    if stream_name in ("engage", "export") and str(properties_flag).lower() == "true":
        schema["additionalProperties"] = True
    else:
        schema["additionalProperties"] = False

    if stream_name == "engage":
        properties = client.request(
            method="GET",
            url=f"https://{client.__api_domain}/api/2.0",
            path="engage/properties",
            params={"limit": 2000},
            endpoint="engage_properties",
        )
        if properties.get("status") == "ok":
            results = properties.get("results", {})
            for key, val in results.items():
                if key[0:1] == "$":
                    new_key = f"mp_reserved_{key[1:]}"
                else:
                    new_key = key

                # property_type: string, number, boolean, datetime, object, list
                # Reference:
                # https://help.mixpanel.com/hc/en-us/articles/115004547063-Properties-Supported-Data-Types
                property_type = val.get("type")

                types = {
                    "boolean": {
                        "type": ["null", "boolean"]},
                    "number": {
                        "type": ["null", "string"],
                        "format": "singer.decimal"},
                    "datetime": {
                        "type": ["null", "string"],
                        "format": "date-time"},
                    "object": {
                        "type": ["null", "object"],
                        "additionalProperties": True,
                    },
                    "list": {
                        "type": ["null", "array"],
                        "required": False,
                        "items": {}},
                    "string": {
                        "type": ["null", "string"]},
                }

                if property_type in types:
                    # Make the types a list containing all types starting with the one returned to us by the API
                    this_type = [types.pop(property_type)]
                    this_type += list(types.values())

                else:
                    this_type = list(types.values())

                schema["properties"][new_key] = {"anyOf": this_type}

    if stream_name == "export":
        # Event properties endpoint:
        #  https://developer.mixpanel.com/docs/data-export-api#section-hr-span-style-font-family-courier-top-span
        results = client.request(
            method="GET",
            url=f"https://{client.__api_domain}/api/2.0",
            path="events/properties/top",
            params={"limit": 2000},
            endpoint="event_properties",
        )
        for key, val in results.items():
            if key[0:1] == "$":
                new_key = f"mp_reserved_{key[1:]}"
            else:
                new_key = key

            # String ONLY for event properties (no other datatypes)
            # Reference: https://help.mixpanel.com/hc/en-us/articles/360001355266-Event-Properties#field-size-character-limits-for-event-properties
            schema["properties"][new_key] = {"type": ["null", "string"]}

    return schema


def get_schemas(client, properties_flag):
    """Load the schema references, prepare metadata for each streams and return
    schema and metadata for the catalog.

    Args:
        client (MixpanelClient): Client object to make http calls.
        properties_flag (bool): Setting this argument to true ensures that new properties on
                                   events and engage records are captured.

    Returns:
        tuple: Returns tuple of Schemas and metadata.
    """
    schemas = {}
    field_metadata = {}

    for stream_name, stream_metadata in STREAMS.items():
        # When the client detects disable_engage_endpoint, skip discovering the stream
        if stream_name == "engage" and client.disable_engage_endpoint:
            LOGGER.warning(
                "Mixpanel returned a 402 indicating the Engage endpoint and stream is unavailable. Skipping."
            )
            continue

        try:
            schema = get_schema(client, properties_flag, stream_name)
        except MixpanelPaymentRequiredError:
            LOGGER.warning(
                "Mixpanel returned a 402 from the %s API so %s stream will be skipped.",
                stream_name,
                stream_name,
            )
            continue

        schemas[stream_name] = schema
        mdata = metadata.new()

        # Documentation:
        # https://github.com/singer-io/getting-started/blob/master/docs/DISCOVERY_MODE.md#singer-python-helper-functions
        # Reference:
        # https://github.com/singer-io/singer-python/blob/master/singer/metadata.py#L25-L44
        mdata = metadata.get_standard_metadata(
            schema=schema,
            key_properties=stream_metadata.key_properties,
            valid_replication_keys=stream_metadata.replication_keys,
            replication_method=stream_metadata.replication_method,
        )

        mdata = metadata.to_map(mdata)

        if stream_metadata.replication_keys:
            mdata = metadata.write(
                mdata,
                ("properties",
                 stream_metadata.replication_keys[0]),
                "inclusion",
                "automatic",
            )

        mdata = metadata.to_list(mdata)

        field_metadata[stream_name] = mdata

    return schemas, field_metadata



import unittest
from unittest import mock

from parameterized import parameterized

from tap_mixpanel.__init__ import main

CONFIG = {
    "api_secret": "dummy_secret",
    "date_window_size": "30",
    "attribution_window": "5",
    "project_timezone": "Europe/Amsterdam",
    "select_properties_by_default": "true",
    "start_date": "2020-02-01T00:00:00Z",
    "user_agent": "tap-mixpanel <api_user_email@your_company.com>",
    "eu_residency": False,
    "end_date": "2020-03-02T00:00:00Z",
}

REQUEST_TIMEOUT_DEFAULT = 300
TIMEOUT_FLOAT = 200.0
TIMEOUT_INT = 200
NULL_STRING = ""
ZERO_INT = 0
ZERO_STRING = "0"
STRING_INT = "200"


class MockParseArgs:
    """Mocked MockParseArgs class with custom state, discover, config
    attributes to pass unit test cases."""

    def __init__(self, state, discover, config):
        self.state = state
        self.discover = discover
        self.config = config


class Mockresponse:
    """Mocked standard HTTPResponse."""

    def __init__(self, resp, status_code):
        self.json_data = resp
        self.status_code = status_code


HEADER = {
    "User-Agent": "tap-mixpanel <api_user_email@your_company.com>",
    "Accept": "application/json",
    "Authorization": "Basic ZHVtbXlfc2VjcmV0",
}


@mock.patch("requests.Session.request", return_value=Mockresponse("", status_code=200))
@mock.patch("singer.utils.parse_args")
@mock.patch("tap_mixpanel.__init__.do_discover", return_value="")
class TestMixpanelRequestTimeoutParameterValue(unittest.TestCase):
    """
    Test that tap handles different type of request_timeout parameter
    values.
    """

    def test_request_timeout_for_none_param_value(
        self, mock_discover, mock_parse_args, mock_request
    ):
        """Test that tap handles none value of request_timeout parameter."""
        config = CONFIG.copy()
        mock_parse_args.return_value = MockParseArgs(
            state={}, discover=True, config=config
        )
        main()

        # Verify that request method called with expected parameter value when"request_timeout" is None
        mock_request.assert_called_with(
            "GET",
            "https://mixpanel.com/api/2.0/engage",
            allow_redirects=True,
            headers=HEADER,
            timeout=REQUEST_TIMEOUT_DEFAULT,
        )

    @parameterized.expand(
        [
            ["empty value", NULL_STRING, REQUEST_TIMEOUT_DEFAULT],
            ["string value", STRING_INT, TIMEOUT_FLOAT],
            ["integer value", TIMEOUT_INT, TIMEOUT_FLOAT],
            ["float value", TIMEOUT_FLOAT, TIMEOUT_FLOAT],
            ["zero value", ZERO_INT, REQUEST_TIMEOUT_DEFAULT],
            ["zero(string) value", ZERO_STRING, REQUEST_TIMEOUT_DEFAULT],
        ]
    )
    def test_request_timeout(
        self, mock_discover, mock_parse_args, mock_request, test_name, input_value, expected_value
    ):
        """Test that tap handles various request timeout values."""
        config = CONFIG.copy()
        config["request_timeout"] = input_value
        mock_parse_args.return_value = MockParseArgs(
            state={}, discover=True, config=config
        )
        main()

        # Verify that request method called with expected parameter value
        mock_request.assert_called_with(
            "GET",
            "https://mixpanel.com/api/2.0/engage",
            allow_redirects=True,
            headers=HEADER,
            timeout=expected_value,
        )
